import numpy as np
#import pyximport; pyximport.install()
import improve_speed


class Game(object):
    """Generic game class that acts as a template"""

    def compute_legal_actions(self, state, player):
        """Return all legal actions for the player in the current state

        Args:
            state: current state
            player: player to return actions for

        Raises:
            NotImplementedError: Description
            NotImplementedError
        """
        raise NotImplementedError()

    def generate_next_state(self, state, action):
        """Returns the state that results from taking a certain action at a certain state

        Args:
            state : initial state
            action : transition action

        Raises:
            NotImplementedError
        """
        raise NotImplementedError()

    def evaluation_function(self, state, player):
        """Returns the desirability of a state for a player

        Args:
            state : state to evaluate
            player : player to evaluate for

        Raises:
            NotImplementedError
        """
        raise NotImplementedError()

    def find_winner(self, state):
        """
        Args:
            state : state to evaluate

        Raises:
            NotImplementedError
        """
        raise NotImplementedError()


class TicTacToe(Game):
    """The classic game of tic tac toe.

    Attributes:
        start_state : blank state
    """
    def __init__(self):
        array = [[" " for i in range(3)] for i in range(3)]
        self.start_state = np.array(array)

    def compute_legal_actions(self, state, player):
        next_actions = []
        for (x, y), value in np.ndenumerate(state):
            if value == " ":
                next_actions.append(((x, y), player))
        return next_actions

    def generate_next_state(self, state, action):
        next_state = np.copy(state)
        position, player = action
        x, y = position
        next_state[x][y] = player
        return next_state

    def get_next_player(self, player):
        if player == "X":
            return "O"
        elif player == "O":
            return "X"
        else:
            print("Not a valid player!!")
            return None

    def convert_state_to_hashable(self, state):
        return ",".join(",".join(a) for a in state)

    def evaluation_function(self, state, player):
        sum = 0
        opponent = self.get_next_player(player)
        for position, owner in np.ndenumerate(state):
            if owner == player:
                sum += self.position_score(position)
            elif owner == opponent:
                sum -= self.position_score(position)

        if self.find_winner(state) == player:
            sum += 1000
        if self.find_winner(state) == opponent:
            sum -= 1000

        return sum

    def is_terminal_state(self, state):
        return self.find_winner(state) is not None or (state == " ").sum() == 0

    def find_winner(self, state):
        for player in ["X", "O"]:
            if (state.diagonal() == player).all() or (np.rot90(state).diagonal() == player).all():
                return player
            for i in range(3):
                if (state[i] == player).all() or (state.T[i] == player).all():
                    return player
        return None

    def position_score(self, position):
        if position == (1, 1):
            return 20
        elif 1 not in position:
            return 5
        else:
            return 2


class Othello(Game):
    """The game of othello.

    Attributes:
        start_state (numpy.ndarray): initial board
    """
    def __init__(self):

        # initialize the board
        start = np.array([[" " for i in range(8)] for i in range(8)])
        start[3][3] = "W"
        start[3][4] = "B"
        start[4][3] = "B"
        start[4][4] = "W"
        self.start_state = start

        # initalize directions as vectors
        directions = []
        for x in range(-1, 2):
            for y in range(-1, 2):
                if x != 0 or y != 0:
                    directions.append(np.array([x, y]))
        self._directions = directions

        # classify positions into categories for later evaluation
        self._corners = [(0, 0), (0, 7), (7, 0), (7, 7)]
        adjacents = []
        diagonals = []
        for x in [0, 1, 6, 7]:
            for y in [0, 1, 6, 7]:
                if x != y:
                    adjacents.append((x, y))
                elif (x, y) not in self._corners:
                    diagonals.append((x, y))
        self._corner_adjacents = adjacents
        self._corner_diagonals = diagonals
        self._side_edges = [(0, 2), (0, 5), (2, 0), (2, 7), (5, 0), (5, 7), (7, 2), (7, 5)]
        self._side_middle = [(0, 3), (0, 4), (3, 0), (3, 7), (4, 0), (4, 7), (7, 3), (7, 4)]
        self._center_corners = [(2, 2), (5, 2), (2, 5), (5, 5)]
        self._center_surroundings = [(2, 3), (2, 4), (3, 2), (3, 5), (4, 2), (4, 5), (5, 3), (5, 4)]
        self._initial_positions = [(3, 3), (3, 4), (4, 3), (4, 4)]

    def compute_legal_actions(self, state, player):
        next_actions = []
        for (x, y), value in np.ndenumerate(state):
            if value == " ":
                for direction in self._directions:
                    if self.is_flippable_in_direction(state, (x, y), player, direction):
                        next_actions.append(((x, y), player))
                        break
        return next_actions

    def generate_next_state(self, state, action):
        next_state = np.copy(state)
        position, player = action
        opponent = self.get_next_player(player)
        x, y = position

        # place the player's stone on the position
        next_state[x][y] = player

        # flip all stones in relevant directions
        for direction in self._directions:
            cur_vector = np.array(position)
            if self.is_flippable_in_direction(state, position, player, direction):
                cur_vector += direction
                x, y = cur_vector
                while self.position_in_bounds(cur_vector) and state[x][y] == opponent:
                    next_state[x][y] = player
                    cur_vector += direction
                    x, y = cur_vector
        return next_state

    def is_flippable_in_direction_slow(self, state, position, player, direction):
        opponent = self.get_next_player(player)
        cur_vector = np.array(position) + direction
        x, y = cur_vector
        if self.position_in_bounds(cur_vector) and state[x][y] == opponent:
            cur_vector += direction
            while self.position_in_bounds(cur_vector) and state[x][y] == opponent:
                x, y = cur_vector
                if state[x][y] == player:
                    return True
                cur_vector += direction
        return False

    def is_flippable_in_direction(self, state, position, player, direction):
        x, y = position
        dx, dy = direction
        return improve_speed.cython_is_flippable_in_direction(state, int(x), int(y), player, int(dx), int(dy))

    def position_in_bounds(self, position):
        x, y = position
        return max(x, y) < 8 and min(x, y) > -1

    def convert_state_to_hashable(self, state):
        return ",\n".join(", ".join(a) for a in state)

    def get_next_player(self, player):
        return "B" if player == "W" else "W"

    def is_terminal_state(self, state):
        if (state == " ").sum() == 0 or (state == "B").sum() == 0 or (state == "W").sum() == 0:
            return True
        elif len(self.compute_legal_actions(state, "B")) + len(self.compute_legal_actions(state, "W")) == 0:
            return True
        else:
            return False

    def find_winner(self, state):
        if self.is_terminal_state(state):
            black_count = (state == "B").sum()
            white_count = (state == "W").sum()
            if black_count > white_count:
                return "black"
            elif white_count > black_count:
                return "white"
            else:
                return "draw"
        else:
            return None

    def evaluation_function(self, state, player):
        opponent = self.get_next_player(player)
        if self.find_winner(state) == player:
            return 10000
        if self.find_winner(state) == opponent:
            return 10000

        pos_score = 0
        for position, value in np.ndenumerate(state):
            if value == player:
                pos_score += self.compute_position_score(position)

        #num_legal_moves = len(self.compute_legal_actions(state, player))
        # turn = (64 - (state == " ").sum()) / 2
        return pos_score# + num_legal_moves * 10

    def improved_evaluation_function(self, state, player, weight_vector={
            "positions": [100, -40, -80, 20, 5, 5, 1, 0, -1],
            "total_stones": 20,
            "legal_moves": 10}):
        opponent = self.get_next_player(player)
        if self.find_winner(state) == player:
            return 10000
        if self.find_winner(state) == opponent:
            return 10000

        # calculate score based on position
        pos_score = 0
        for position, value in np.ndenumerate(state):
            if value == player:
                pos_score += self.compute_position_score(position, weights=weight_vector["positions"])

        num_legal_moves = len(self.compute_legal_actions(state, player))
        num_stones = (state == player).sum()

        # change the evaluation function based on the stage of the game
        turn = 64 - (state == " ").sum()
        if turn > 43:
            return pos_score + num_stones * weight_vector["total_stones"] + num_legal_moves
        else:
            return pos_score + num_legal_moves * weight_vector["legal_moves"]

    def compute_position_score(self, position, weights=[100, -40, -80, 20, 5, 5, 1, 0, -1]):

        """
        self._corners = [(0, 0), (0, 7), (7, 0), (7, 7)]
        adjacents = []
        diagonals = []
        for x in [0, 1, 6, 7]:
            for y in [0, 1, 6, 7]:
                if x != y:
                    adjacents.append((x, y))
                elif (x, y) not in self._corners:
                    diagonals.append((x, y))
        self._corner_adjacents = adjacents
        self._corner_diagonals = diagonals
        self._side_edges = [(0, 2), (0, 5), (2, 0), (2, 7), (5, 0), (5, 7), (7, 2), (7, 5)]
        self._side_middle = [(0, 3), (0, 4), (3, 0), (3, 7), (4, 0), (4, 7), (7, 3), (7, 4)]
        self._center_corners = [(2, 2), (5, 2), (2, 5), (5, 5)]
        self._center_surroundings = [(2, 3), (2, 4), (3, 2), (3, 5), (4, 2), (4, 5), (5, 3), (5, 4)]
        self._initial_positions = [(3, 4), (3, 4), (4, 3), (4, 4)]

        """

        if position in self._corners:
            return weights[0]
        elif position in self._corner_adjacents:
            return weights[1]
        elif position in self._corner_diagonals:
            return weights[2]
        elif position in self._side_edges:
            return weights[3]
        elif position in self._side_middle:
            return weights[4]
        elif position in self._center_corners:
            return weights[5]
        elif position in self._center_surroundings:
            return weights[6]
        elif position in self._initial_positions:
            return weights[7]
        else:
            return weights[8]

    def fools_mate_evaluation_function(self, state, player, weights=[1, 0, 0, 0, 0, 0, 1]):
        num_black = (state == "B").sum()
        num_white = (state == "W").sum()
        if num_white == 0:
            return 10000000
        num_black_legal_moves = len(self.compute_legal_actions(state, "B"))
        num_white_legal_moves = len(self.compute_legal_actions(state, "W"))
        pos_score = 0
        for position, value in np.ndenumerate(state):
            if value != " ":
                pos_score += self.compute_position_score(position, weights=[-100, -100, -100, -100, -100, 0, 0, 0, -1])

        # compute number of regions
        num_white_regions = improve_speed.cython_compute_num_regions(state, "W")
        num_black_regions = improve_speed.cython_compute_num_regions(state, "B")

        # compute average number of adjacent black stones for each white stone
        num_adjacent_black = compute_adjacent_black_stones(state) / num_white

        score = 0
        score += num_black * weights[0]
        score += -1 * num_white * weights[1]
        score += -1 * num_white_regions * weights[2]
        score += -1 * num_black_legal_moves * weights[3]
        score += num_white_legal_moves * weights[4]
        score += num_adjacent_black * weights[5]
        score += pos_score * weights[6]
        if len(weights) > 7:
            score += num_black_regions * weights[7]

        return score

    def fast_foolsmate_evaluation_function(self, state, player):
        num_white = (state == "W").sum()
        if num_white == 0:
            return 10000000
        pos_score = 0
        num_white_regions = improve_speed.cython_compute_num_regions(state, "W")
        for position, value in np.ndenumerate(state):
            if value != " ":
                pos_score += self.compute_position_score(position, weights=[-100, -100, -100, -100, -100, 0, 0, 0, -1])

        return pos_score - num_white - num_white_regions


def compute_adjacent_black_stones(state):
    directions = []
    for x in range(-1, 2):
        for y in range(-1, 2):
            if x != 0 or y != 0:
                directions.append(np.array([x, y]))
    num_adjacent_black = 0
    for position, value in np.ndenumerate(state):
            if value == "W":
                for direction in directions:
                    cur_vector = np.array(position) + direction
                    x, y = cur_vector
                    if x > 0 and y > 0 and x < 8 and y < 8 and state[x, y] == "B":
                        num_adjacent_black += 1
    return num_adjacent_black
