import search
import games
import argparse
import sys


parser = argparse.ArgumentParser(
    description="""Game player.
    """, formatter_class=argparse.RawTextHelpFormatter)

parser.add_argument("-g", "--game", type=str,
    default="othello", help="""The game you want to play. Options are
        - tictactoe
        - othello
        - foolsmate
    """)


parser.add_argument("-l", "--log", type=str,
    default="", help="""The file to log output to.""")

global divide


def playout(game, player1, player2):
    state = game.start_state
    turn_player = player1

    print("Game Starting...")
    print(state)
    while not game.is_terminal_state(state):
        print(divide)
        action = turn_player.compute_action(game, state)
        if action is None:
            print("No actions available...")
        else:
            print("Action: {}".format(action))
            state = game.generate_next_state(state, action)
        print(state)
        turn_player = player2 if turn_player.player == player1.player else player1

    print(divide)
    print("Results:")
    print("{}: {}".format(player1.player, (state == player1.player).sum()))
    print("{}: {}".format(player2.player, (state == player2.player).sum()))
    print("Winner is {}".format(game.find_winner(state)))
    return (state == player1.player).sum(), (state == player2.player).sum()


def tictactoe():
    global divide
    divide = "-" * 20
    game = games.TicTacToe()
    player1 = search.TreeSearchPlayer(2, 'X', game.evaluation_function)
    player2 = search.TreeSearchPlayer(2, 'O', game.evaluation_function)

    playout(game, player1, player2)
    print(player1.expanded_nodes)
    print(player2.expanded_nodes)


def othello(cache=False):
    global divide
    divide = "-" * 35
    game = games.Othello()
    player1 = search.TreeSearchPlayer(3, 'B', game.improved_evaluation_function, cache=cache)
    player2 = search.TreeSearchPlayer(3, 'W', game.improved_evaluation_function, cache=cache)

    playout(game, player1, player2)

    print("Expanded Nodes: player1 = {}, player2 = {}".format(player1.expanded_nodes, player2.expanded_nodes))


def check_othello_performance(depth, logfile=""):
    global divide

    divide = "-" * 35
    game = games.Othello()
    player1 = search.TreeSearchPlayer(depth, 'B', game.evaluation_function, cache=False)
    player2 = search.RandomPlayer('W')

    score1, score2 = playout(game, player1, player2)

    print("Expanded Nodes: player1 = {}".format(player1.expanded_nodes))

    if len(logfile) > 0:
        with open(logfile, "a") as file:
            file.write("{}, {}, {}, {}\n".format(depth, score1, score2, player1.expanded_nodes))


def foolsmate():
    global divide
    divide = "-" * 35
    game = games.Othello()
    player1 = search.TreeSearchCooperativePlayer(2, "B", game.fast_foolsmate_evaluation_function, cache=True)
    player2 = search.TreeSearchCooperativePlayer(2, "W", game.fast_foolsmate_evaluation_function, cache=True)

    playout(game, player1, player2)

    print("Expanded Nodes: player1 = {}, player2 = {}".format(player1.expanded_nodes, player2.expanded_nodes))


def measure_performance(weights, logfile=""):
    game = games.Othello()
    num_weights = [float(s) for s in args.options.split(":")]
    player1 = search.TreeSearchCooperativePlayer(1, "B", lambda x, y: game.fools_mate_evaluation_function(x, y, weights=num_weights), cache=False)
    player2 = search.TreeSearchCooperativePlayer(1, "W", lambda x, y: game.fools_mate_evaluation_function(x, y, weights=num_weights), cache=False)

    state = game.start_state
    turn_player = player1
    turn = 0
    while not game.is_terminal_state(state):
        turn += 1
        if turn > 9:
            break
        action = turn_player.compute_action(game, state)
        if action is not None:
            state = game.generate_next_state(state, action)
        turn_player = player2 if turn_player.player == player1.player else player1

    if len(logfile) > 0:
        with open(logfile, "a") as file:
            file.write(",".join(weights) + "," + str(turn) + "\n")


if __name__ == '__main__':
    args = parser.parse_args()
    game = args.game
    if game == "tictactoe":
        tictactoe()
    elif game == "othello":
        othello()
    elif game == "foolsmate":
        foolsmate()
    else:
        print("Sorry, that game does not exist. Consult help to see options available!")
        sys.exit(1)
