import random


def log_eval_function(func):
    def wrapper(*args, **kwargs):
        value, action = func(*args, **kwargs)
        with open("eval_values.csv", "a") as file:
            file.write("{}\n".format(value))
        return value, action
    return wrapper


class TreeSearchPlayer(object):
    """A player that uses tree search (minmax or alpha beta) to play."""

    def __init__(self, max_depth, player, evaluation_function, alphabeta=True, cache=True):
        self.max_depth = max_depth
        self.player = player
        self.expanded_nodes = 0
        self.evaluation_function = evaluation_function
        self.calculated_values = {}
        self.use_alphabeta = alphabeta
        self.cache = cache

    def compute_action(self, game, state):
        value, action = self.compute_best_action_and_value(game, state, self.player, -1, -float("inf"), float("inf"))
        return action

    @log_eval_function
    def compute_best_action_and_value(self, game, state, player, depth, alpha, beta):
        self.expanded_nodes += 1

        my_turn = player == self.player
        depth += 1

        actions = game.compute_legal_actions(state, player)

        if depth == self.max_depth or game.is_terminal_state(state) or len(actions) == 0:
            value = self.evaluation_function(state, self.player)
            return (value, None)

        if my_turn:
            max_value = -float("inf")
            max_action = None
            for action in actions:
                child_state = game.generate_next_state(state, action)
                hashed_state = game.convert_state_to_hashable(child_state)
                if child_state is not None:
                    if self.cache and hashed_state in self.calculated_values and self.calculated_values[hashed_state]:
                        child_value = self.calculated_values[hashed_state]
                    else:
                        child_value = self.compute_best_action_and_value(game, child_state, game.get_next_player(player), depth, alpha, beta)[0]
                        self.calculated_values[hashed_state] = child_value
                    if child_value > max_value:
                        max_value = child_value
                        max_action = action

                    # conduct beta cut, since this branch will never be chosen anyway
                    if self.use_alphabeta:
                        alpha = max(alpha, max_value)
                        if max_value > beta:
                            return (max_value, max_action)
            return (max_value, max_action)
        else:
            min_value = float("inf")
            min_action = None
            for action in actions:
                child_state = game.generate_next_state(state, action)
                hashed_state = game.convert_state_to_hashable(child_state)
                if child_state is not None:
                    if self.cache and hashed_state in self.calculated_values and self.calculated_values[hashed_state]:
                        child_value = self.calculated_values[hashed_state]
                    else:
                        child_value = self.compute_best_action_and_value(game, child_state, game.get_next_player(player), depth, alpha, beta)[0]
                        if self.cache:
                            self.calculated_values[hashed_state] = child_value
                    if child_value < min_value:
                        min_value = child_value
                        min_action = action

                    # conduct alpha cut, since this branch will never be chosen anyway
                    if self.use_alphabeta:
                        beta = min(beta, min_value)
                        if min_value < alpha:
                            return (min_value, min_action)

            return (min_value, min_action)


class TreeSearchCooperativePlayer(object):
    """A cooperative player that uses tree search to play."""

    def __init__(self, max_depth, player, evaluation_function, cache=False):
        self.max_depth = max_depth
        self.player = player
        self.expanded_nodes = 0
        self.evaluation_function = evaluation_function
        self.calculated_values = {}
        self.cache = cache

    def compute_action(self, game, state):
        value, action = self.compute_best_action_and_value(game, state, self.player, -1)
        return action

    def compute_best_action_and_value(self, game, state, player, depth):
        self.expanded_nodes += 1

        my_turn = player == self.player
        depth += 1

        actions = game.compute_legal_actions(state, player)

        if depth == self.max_depth or game.is_terminal_state(state) or len(actions) == 0:
            value = self.evaluation_function(state, self.player)
            return (value, None)

        max_value = -float("inf")
        max_action = None
        for action in actions:
            child_state = game.generate_next_state(state, action)
            hashed_state = game.convert_state_to_hashable(child_state)
            if child_state is not None:
                if self.cache and hashed_state in self.calculated_values and self.calculated_values[hashed_state]:
                    child_value = self.calculated_values[hashed_state]
                else:
                    child_value = self.compute_best_action_and_value(game, child_state, game.get_next_player(player), depth)[0]
                    self.calculated_values[hashed_state] = child_value
                if child_value > max_value:
                    max_value = child_value
                    max_action = action

        return (max_value, max_action)


class RandomPlayer(object):

    def __init__(self, player):
        self.player = player

    def compute_action(self, game, state):
        actions = game.compute_legal_actions(state, self.player)
        if len(actions) == 0:
            return None
        else:
            return random.choice(game.compute_legal_actions(state, self.player))
